//---------------------------------------------------------------------------
// Copyright (C) 2000 Dallas Semiconductor Corporation, All Rights Reserved.
// 
// Permission is hereby granted, free of charge, to any person obtaining a 
// copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included 
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES 
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
// OTHER DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of Dallas Semiconductor 
// shall not be used except as stated in the Dallas Semiconductor 
// Branding Policy. 
//---------------------------------------------------------------------------
//
//  ibshasc.C - SHA-iButton script functions
//
//  Compiler:  Microsoft Visual C 
//  Version:   1.02
//  History:   
//
//  Debug code designated with      //\\//\\//\\//\\//


#include "ibtmexcw.h"   
#include "ibsha.h"

// external function prototypes
extern int ReadFromMIB(ushort, uchar *, int);
extern void output_status(int, char *);
extern int EraseScratchSHA(void);
extern int ReadScratchSHA(ushort *, uchar *, uchar *);
extern int WriteScratchSHA(ushort, uchar *, int);
extern int CopyScratchSHA(ushort,int);
extern int ReadAuthPageSHA(ushort, uchar *);
extern int ReadPageUDP(ushort, uchar *);
extern int ComputeSHA(ushort, uchar);
extern int MatchScratchSHA(uchar *);
extern ushort docrc16(ushort);
extern int FormatUDP(ushort, uchar *, int *);
extern int ReadUDP_SHA(ushort address, uchar *, int *);
extern int ReadFileSHA(uchar *, int *, uchar *, int *, int);
extern int TestModeMatch(void); //\\//\\//\\//

// Local Function Prototypes 
int RunScriptSHA(Script script[], void *, int);
int ScriptStepSHA(Script *, int *, int *, int *, char *, void *, int);
int Make_eCertificate(ushort, uchar *, eCertificate *);

// globals
extern ushort CRC16;

// defines
#define MAX_TRANS_ERROR  5

//--------------------------------------------------------------------------
// Script functions
//--------------------------------------------------------------------------

//--------------------------------------------------------------------------
// Run the specified script. 
// Status is printed using 'output_status' function.  The script result is
// returned in the buffer 'prslt' of max length 'rslt_max'.  Result data 
// format is script dependent.
//
// Return: TRUE - all steps completed
//         FALSE - all steps NOT completed due to error or escape 
//
int RunScriptSHA(Script script[], void *prslt, int rslt_max)
{
   char statusmsg[256],LastDescription[256],LastMsg[256];
   char tmsg[256];
   int StepCount,SubStep,ErrorCount,Status;
   uchar *puchar;   

   static ulong st = 0;

   // reset the step to the begining
   StepCount = 0;
   SubStep = 0;
   ErrorCount = 0;
   Status = STATUS_INPROGRESS;
   statusmsg[0] = 0;
   puchar = prslt;
 
   // start of script marker
   output_status(LV_VERBOSE,"-------------------------------------\n");

   // loop to perform all of the steps to download the Thermochron
   do
   {   
      // switch on the status of the last step done
      switch(Status)
      {
         // step complete so go to the next
         case STATUS_STEP_COMPLETE:
            StepCount++;
            SubStep = 0;
            ErrorCount = 0;
            Status = STATUS_INPROGRESS;
            LastDescription[0] = 0;
            LastMsg[0] = 0;
            break;
         // in progress so call again
         case STATUS_INPROGRESS:
            // print step description if different 
            if (strcmp(LastDescription,
                script[StepCount].StepDescription) != 0)
            {
               sprintf(tmsg,"(%d)%s -->",StepCount,script[StepCount].StepDescription);
               output_status(LV_VERBOSE,tmsg);
               sprintf(LastDescription,"%s",script[StepCount].StepDescription);
            }

            // perform a step in the job
            ScriptStepSHA(&script[StepCount],&SubStep, 
                      &Status, &ErrorCount, statusmsg, 
                      prslt, rslt_max);

            // print results if different
            if (strcmp(LastMsg,statusmsg) != 0)
            {
               sprintf(tmsg,"%s\n",statusmsg); 
               output_status(LV_VERBOSE,tmsg);
               sprintf(LastMsg,"%s",statusmsg);
            }
            else
               output_status(LV_VERBOSE,".");
            break;
         // encountered a transient error
         case STATUS_ERROR_TRANSIENT:
            // if MAX_TRANS_ERROR tansient errors in a row then abort
            if (ErrorCount > MAX_TRANS_ERROR)
               Status = STATUS_ERROR_HALT;
            else
               Status = STATUS_INPROGRESS;
            break;
         // all steps complete
         case STATUS_COMPLETE:
            output_status(LV_VERBOSE,"End script normally\n");
            return TRUE;
         // non-recoverable error
         case STATUS_ERROR_HALT:
            // try to do emergency clean up
            output_status(LV_VERBOSE,"Aborting script due to non-recoverable error\n");
            return FALSE;
      }
   }
   while (!kbhit());

   // key abort
   output_status(LV_ALWAYS,"Aborting script due to key press\n");
   return FALSE;
}

//----------------------------------------------------------------------
//  Use the script to perform a step and return.
//
//  Return:
//          STATUS_STEP_COMPLETE - step complete, go to next step
//          STATUS_INPROGRESS - step is not complete, call again
//          STATUS_ERROR_HALT -  error, halt script
//          STATUS_ERROR_TRANSIENT - transient error, call again
//
int ScriptStepSHA(Script *StateScript, int *SubStep,
              int *Status, int *ErrorCount, char *statusmsg,
              void *prslt, int rslt_max)
{
   int rslt,statuslen,j,i,cnt;
   short *pshort;
   short ROM[8];
   ulong tcnt;
   eCertificate *ec;
   uchar *puchar;

   static ushort address;
   static uchar es, sha_cmd;
   static ulong target_secret; 
   static uchar randnum[3];
   static uchar temp_buffer[64];
   static int buf_len;
   
   // do the current step and return
   switch (StateScript->Step)
   {
      // attempt to get a session on the 1-Wire Net
      case ST_GET_SESSION:
         // check for same port/type as other device already has session
         if ((PortNum[IB_COPR] == PortNum[IB_ROVING]) && 
             (PortType[IB_COPR] == PortType[IB_ROVING]))
         {
            if (dev == IB_COPR)
               rslt = TMValidSession(session_handle[IB_ROVING]);
            else
               rslt = TMValidSession(session_handle[IB_COPR]);
      
            if (rslt >= 1)
            {
               if (dev == IB_COPR)
                  session_handle[IB_COPR] = session_handle[IB_ROVING];
               else
                  session_handle[IB_ROVING] = session_handle[IB_COPR];
               sprintf(statusmsg,"Session aquired");
               *Status = STATUS_STEP_COMPLETE;
               break;
            }
         }
         // must not already have a session so continue
         session_handle[dev] = TMExtendedStartSession(PortNum[dev],PortType[dev],NULL);
         if (session_handle[dev] > 0)
         {
            sprintf(statusmsg,"Session aquired");
            *Status = STATUS_STEP_COMPLETE;
         }
         else if (session_handle[dev] == 0)
         {
            sprintf(statusmsg,"1-Wire Net in use by other process");
            *Status = STATUS_INPROGRESS;
         }
         else if (session_handle[dev] < 0)
         {
            sprintf(statusmsg,"1-Wire Net not available");
            *Status = STATUS_ERROR_HALT;
         }
         break;
      
      // perform TMSetup on the 1-Wire Net port   
      case ST_SETUP:
         rslt = TMSetup(session_handle[dev]);
         if (rslt == 1)
         {
            sprintf(statusmsg,"1-Wire setup is success");
            *Status = STATUS_STEP_COMPLETE;
         }
         else if (rslt == 2)
         {
            sprintf(statusmsg,"1-Wire shorted");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         else
         {
            sprintf(statusmsg,"1-Wire Net has a hardware problem, abort");
            *Status = STATUS_ERROR_HALT;
         }
         break;

      // release the session on the 1-Wire Net
      case ST_REL_SESSION:
         TMEndSession(session_handle[dev]);
         sprintf(statusmsg,"1-Wire Net released");
         *Status = STATUS_STEP_COMPLETE;
         break;      

      // the operation is complete      
      case ST_FINISH:
         sprintf(statusmsg,"Operation complete");
         *Status = STATUS_COMPLETE;
         break; 

      // set the search to target the SHA iButton
      case ST_SEARCH_TARGET:
         TMFamilySearchSetup(session_handle[dev],&state_buffer[dev][0],SHA_FAM);
         sprintf(statusmsg,"Search setup to look for SHA iButton");
         // force back to standard speed
         if (TMOneWireCom(session_handle[dev],1,0) == 1)
            TMOneWireCom(session_handle[dev],0,0);
         *Status = STATUS_STEP_COMPLETE;
         break; 

      // search to find a SHA, verify is a SHA and not EQUAL to copr_rom[]
      case ST_FIND_SHA:
         for (;;)
         {
            // get first device in list with the specified family 
            if (TMNext(session_handle[dev],&state_buffer[dev][0]) == 1)
            {
               // verify correct device type
               ROM[0] = 0;
               TMRom(session_handle[dev],&state_buffer[dev][0],ROM);   

               // compare to copr_rom
               for (i = 0; i < 8; i++)
               {
                  if (copr_rom[i] != ROM[i])
                     break;
               }

               // check if correct type and not copr_rom 
               if ((SHA_FAM == (ROM[0] & 0x7F)) && (i != 8))
               {
                  // save the ROM to the return buffer
                  pshort = prslt;
                  pshort[0] = 0;
                  TMRom(session_handle[dev],&state_buffer[dev][0],pshort);   

                  statuslen = sprintf(statusmsg,"SHA iButton found: ");
                  for (j = 7; j >= 0; j--)
                     statuslen += sprintf(&statusmsg[statuslen],"%02X",ROM[j]);
                  *Status = STATUS_STEP_COMPLETE;
                  break;
               }
            } 
            else
            {
               sprintf(statusmsg,"SHA iButton not on 1-Wire Net");
               *Status = STATUS_ERROR_HALT;
               break;
            }
         }
         break;

      // check if device found is a 'roving' type
      case ST_CHECK_ROVING:
         // read the roving service file
         rslt = ReadFileSHA(provider, &i, temp_buffer, &buf_len, 64);
         if (rslt)
         {
            // look at length to see if have user data
            if (buf_len > 28)
            {
               // extract the user data in global
               for (j = 0; j < (buf_len - 28); j++)
                  user_data[j] = temp_buffer[j];
               user_data[j] = 0;

               // set the money page secret from the start page of file
               r_udevice_scrt = i;
               // stop script short on success
               sprintf(statusmsg,"Roving device found (%d bytes user data)",buf_len-28);
               *Status = STATUS_COMPLETE;  
               break;
            }
            else if (buf_len == 28)
            {
               // set the money page secret from the start page of file
               if (i > 7)
                  r_udevice_scrt = i - 8;
               // stop script short on success
               sprintf(statusmsg,"Roving device found (no user data)");
               *Status = STATUS_COMPLETE;  
               break;
            }
            else
            {
               sprintf(statusmsg,"Roving money file incorrect size");
               *Status = STATUS_ERROR_HALT;
               break;
            }
         }
         sprintf(statusmsg,"Could not read directory");
         // check for error last time 
         *ErrorCount = *ErrorCount + 1;
         *Status = STATUS_ERROR_TRANSIENT;
         // check if close to erroring out of this step
         if (*ErrorCount >= MAX_TRANS_ERROR)
         {
            sprintf(statusmsg,"Giving up on this SHA iButton, no roving service file");
            *ErrorCount = 0;
            *Status = STATUS_STEP_COMPLETE;
         }
         break;
   
      // check if device found is a 'roving' type
      case ST_CHECK_COPROC:
         // read the co-processor information file
         rslt = ReadFileSHA(copr_file, &i, temp_buffer, &buf_len, 64);
         if (rslt)
         {
            // extract the information out of the copr_file
            if (buf_len >= 8)
            {
               // get provider name
               for (i = 0; i < 5; i++)
                  provider[i] = temp_buffer[i];
               // get money master scrt number
               c_mmaster_scrt = temp_buffer[5];
               // auth master secret number
               c_amaster_scrt = temp_buffer[6];
               // unique device secret(calculated)
               c_udevice_scrt = temp_buffer[7];
               // stop script short on success
               sprintf(statusmsg,"Co-processor found");
               *Status = STATUS_COMPLETE;  
               break;
            }
            else
            {
               sprintf(statusmsg,"Co-processor info file incorrect size");
               *Status = STATUS_ERROR_HALT;
               break;
            }
         }
         sprintf(statusmsg,"Could not read directory");
         // check for error last time 
         *ErrorCount = *ErrorCount + 1;
         *Status = STATUS_ERROR_TRANSIENT;
         // check if close to erroring out of this step
         if (*ErrorCount >= MAX_TRANS_ERROR)
         {
            sprintf(statusmsg,"Giving up on this SHA iButton, no co-processor info file");
            *ErrorCount = 0;
            *Status = STATUS_STEP_COMPLETE;
         }
         break;

      // erase the scratch of the token
      case ST_ERASE_SCRATCH:
         if (EraseScratchSHA())
         {
            sprintf(statusmsg,"Scratchpad erased");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            cnt = sprintf(statusmsg,"Could not erase scratchpad, device not present");
            //\\//\\//\\//\\// TRY TEST MODE FIX
            TestModeMatch();
            //\\//\\//\\//\\// END TEST MODE FIX

            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;

      // read the scratch of the token
      case ST_READ_SCRATCH:
         if (ReadScratchSHA(&address,&es,temp_buffer))
         {
            sprintf(statusmsg,"Scratchpad read");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            sprintf(statusmsg,"Could not read scratchpad, device not present");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;
         
      // write the scratch of the token
      case ST_WRITE_SCRATCH:
         // adjust length if too long
         if (buf_len > (0x20 - (address & 0x1F)))
            i = (0x20 - (address & 0x1F));
         else
            i = buf_len;        
         if (WriteScratchSHA(address,temp_buffer,i))
         {
            sprintf(statusmsg,"Scratchpad written");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            sprintf(statusmsg,"Could not write scratchpad, device not present");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;

      // copy the scratch to the token
      case ST_COPY_SCRATCH:
         if (CopyScratchSHA(address,buf_len))
         {
            sprintf(statusmsg,"Scratchpad copy complete");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            sprintf(statusmsg,"Could not copy scratchpad, device not present");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;

      // read authenticate page
      case ST_READ_AUTH_PAGE:
         if (ReadAuthPageSHA(address,temp_buffer))
         {
            sprintf(statusmsg,"Read Authenticate page complete");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            sprintf(statusmsg,"Could not read auth page, device not present");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;

      // try to make an eCertificate from the buffer contents
      case ST_MAKE_ECERT:
         // fill in the eCertificate for return
         if (Make_eCertificate(address,temp_buffer,(eCertificate*)prslt))
         {
            sprintf(statusmsg,"eCertificate successfully made");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            // not valid money page 
            sprintf(statusmsg,"Not valid money page, abort");
            *Status = STATUS_ERROR_HALT;
         }
         break;

      // compute sha
      case ST_COMPUTE_SHA:
         if (ComputeSHA(address, sha_cmd))
         {
            sprintf(statusmsg,"SHA computed");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            sprintf(statusmsg,"Could not compute SHA, device not present");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;

      // match scratch
      case ST_MATCH_SCRATCH_MONEY:
         // save mac 
         if (!prslt)
         {
            sprintf(statusmsg,"eCertificate object not present");
            *Status = STATUS_ERROR_HALT;
            break;
         }
         // get a pointer to the eCert
         ec = (eCertificate*)prslt;
         if (MatchScratchSHA(ec->money_mac))
         {
            sprintf(statusmsg,"Scratchpad matched!");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            sprintf(statusmsg,"Scratchpad did not match");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;

      // match scratch
      case ST_MATCH_SCRATCH_AUTH:
         // save mac 
         if (!prslt)
         {
            sprintf(statusmsg,"eCertificate object not present");
            *Status = STATUS_ERROR_HALT;
            break;
         }
         // get a pointer to the eCert
         ec = (eCertificate*)prslt;
         if (MatchScratchSHA(ec->auth_mac))
         {
            sprintf(statusmsg,"Scratchpad matched!");
            *Status = STATUS_STEP_COMPLETE;
         }
         else
         {
            sprintf(statusmsg,"Scratchpad did not match");
            // check for error last time 
            *ErrorCount = *ErrorCount + 1;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         break;

      // check to see if the current device is present
      case ST_PRESENT:
         if (TMStrongAccess(session_handle[dev],&state_buffer[dev][0]) != 1)
         {
            // not present so try to force out of overdrive
            TMAccess(session_handle[dev],&state_buffer[dev][0]);

            sprintf(statusmsg,"Device Not present");
            // check for error last time 
            *ErrorCount = *ErrorCount + 2;
            *Status = STATUS_ERROR_TRANSIENT;
         }
         else
         {
            sprintf(statusmsg,"Device present");
            *Status = STATUS_STEP_COMPLETE;
         }
         break;

      // select the ROVING
      case ST_SELECT_ROVING:
         dev = IB_ROVING;
         sprintf(statusmsg,"Token is now current device");
         *Status = STATUS_STEP_COMPLETE;
         break;

      // select the CO-PROCESSOR
      case ST_SELECT_COPR:
         dev = IB_COPR;
         sprintf(statusmsg,"CO-Processor is now current device");
         *Status = STATUS_STEP_COMPLETE;
         break;

      // selet the co-processor money master secret
      case ST_TARGET_C_MMS:
         // get the target secret
         target_secret = c_mmaster_scrt;
         sprintf(statusmsg,"Target secret %d",target_secret);
         *Status = STATUS_STEP_COMPLETE;
         break;         

      // select the co-processor auth master secret
      case ST_TARGET_C_AMS:
         // get the target secret
         target_secret = c_amaster_scrt;
         sprintf(statusmsg,"Target secret %d",target_secret);
         *Status = STATUS_STEP_COMPLETE;
         break;         

      // select the co-processor unique device secret(calculated)
      case ST_TARGET_C_UDS:
         // get the target secret
         target_secret = c_udevice_scrt;
         sprintf(statusmsg,"Target secret %d",target_secret);
         *Status = STATUS_STEP_COMPLETE;
         break;         

      // select the roving unique device secret
      case ST_TARGET_R_UDS:
         // get the target secret
         target_secret = r_udevice_scrt;
         sprintf(statusmsg,"Target secret %d",target_secret);
         *Status = STATUS_STEP_COMPLETE;
         break;

      // select the till unique device secret
      case ST_TARGET_T_UDS:
         // get the target secret
         target_secret = t_udevice_scrt;
         sprintf(statusmsg,"Target secret %d",target_secret);
         *Status = STATUS_STEP_COMPLETE;
         break;         

      // select the page associated with the target secret (with counter)
      case ST_TARGET_PAGE:
         // calculate the desired page address
         address = 0x100 + (ushort)(target_secret * 0x20);
         buf_len = 32;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Target address set to %4X (page %d)",address,address / 32);
         break;         

      // select the directory page
      case ST_TARGET_DIRECTORY:
         address = 0;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Target address set to %4X (page %d)",address,address / 32);
         break;         

      // select the page associated with the target secret (with counter)
      case ST_TARGET_SECRET:
         // calculate the desired page address
         address = 0x200 + (ushort)(target_secret * 0x08);
         buf_len = 8;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Target address set to %04X (secret %d)",address,address / 8);
         break;         
   
      // set compute sha command
      case ST_SHA_COMPUTE_FIRST_SCRT:
         sha_cmd = 0x0F;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"SHA command set (%02X)",sha_cmd);
         break;         
      case ST_SHA_COMPUTE_NEXT_SCRT:
         sha_cmd = 0xF0;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"SHA command set (%02X)",sha_cmd);
         break;         
      case ST_SHA_VALIDATE_PAGE:
         sha_cmd = 0x3C;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"SHA command set (%02X)",sha_cmd);
         break;         
      case ST_SHA_SIGN_PAGE:
         sha_cmd = 0xC3;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"SHA command set (%02X)",sha_cmd);
         break;         
      case ST_SHA_COMPUTE_CHALLENGE:
         sha_cmd = 0xCC;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"SHA command set (%02X)",sha_cmd);
         break;         
      case ST_SHA_AUTHENTICATE_HOST:
         sha_cmd = 0xAA;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"SHA command set (%02X)",sha_cmd);
         break;         

      // set the temp buffer to 0xFF's
      case ST_SET_BUFF_FF:
         for (i = 0; i < 32; i++)
            temp_buffer[i] = 0xFF;
         buf_len = 32;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Temporary buffer set to FF's");
         break;         

      // set the temp buffer to 0x00's
      case ST_SET_BUFF_00:
         for (i = 0; i < 32; i++)
            temp_buffer[i] = 0;
         buf_len = 32;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Temporary buffer set to 00's");
         break;         

      // set the buffer to the ROM number
      case ST_SET_BUFF_ROM:
         ROM[0] = 0;
         TMRom(session_handle[dev],&state_buffer[dev][0],ROM);
         for (i = 0; i < 32; i++)
            temp_buffer[i] = 0xFF;
         for (i = 0; i < 7; i++)
            temp_buffer[i+13] = (uchar)ROM[i];   
         buf_len = 32;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Temporary buffer set to ROM at offset 13");
         break;         

      // get a new random challenge value
      case ST_GET_NEW_RAND:
         // seed random number generator with time
         srand((unsigned int)time(NULL)); 
         tcnt = rand(); 
         for (i = 0; i < 3; i++)
         {
            randnum[i] = (uchar)(tcnt & 0xFF); 
            tcnt >>= 8;
         }
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"New random value retrieved");
         break;         

      // set random challenge value to NULL
      case ST_SET_RAND_NULL:
         for (i = 0; i < 3; i++)
            randnum[i] = 0;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Random challenge is NULL");
         break;

      // set random value in 20 to 22 in temp buffer
      case ST_SET_BUFF_RAND:
         for (i = 0; i < 32; i++)
            temp_buffer[i] = 0xFF;
         for (i = 0; i < 3; i++)
            temp_buffer[i+20] = randnum[i]; 
         buf_len = 32;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Temporary buffer set to random value");
         break;         

      // save the MAC in the temp_buffer as the money MAC in the eCert 
      case ST_SAVE_MONEY_MAC:
         // save mac 
         if (prslt)
         {
            ec = (eCertificate*)prslt;
            for (i = 0; i < 20; i++)
               ec->money_mac[i] = temp_buffer[i + 8];
            *Status = STATUS_STEP_COMPLETE;
            sprintf(statusmsg,"Money MAC save in eCertificate");
         }
         else
         {
            sprintf(statusmsg,"eCertificate object not present");
            *Status = STATUS_ERROR_HALT;
         }
         break;

      // save the MAC in the temp_buffer as the auth MAC in the eCert 
      case ST_SAVE_AUTH_MAC:
         // save mac 
         if (prslt)
         {
            ec = (eCertificate*)prslt;
            for (i = 0; i < 20; i++)
               ec->auth_mac[i] = temp_buffer[i + 8];
            *Status = STATUS_STEP_COMPLETE;
            sprintf(statusmsg,"Auth MAC save in eCertificate");
         }
         else
         {
            sprintf(statusmsg,"eCertificate object not present");
            *Status = STATUS_ERROR_HALT;
         }
         break;

      // set account info in temp buffer (money, transaction ID, MAC)
      case ST_SET_BUFF_ACCNT_FXT:
         // save mac 
         if (!prslt)
         {
            sprintf(statusmsg,"eCertificate object not present");
            *Status = STATUS_ERROR_HALT;
            break;
         }
         // get a pointer to the eCert
         ec = (eCertificate*)prslt;
         // length byte
         cnt = 0;
         temp_buffer[cnt++] = 0x1D;
         // transaction ID
         temp_buffer[cnt++] = ec->trans_id & 0xFF;
         temp_buffer[cnt++] = (ec->trans_id >> 8) & 0xFF;
         // monetary unit code & Multiplier
         temp_buffer[cnt++] = ec->code_mult & 0xFF;
         temp_buffer[cnt++] = (ec->code_mult >> 8) & 0xFF;
         // money balance 
         for (i = 0; i < 3; i++)
            temp_buffer[cnt++] = ec->balance[i];
         // MAC (set to all zeros)
         for (i = 0; i < 20; i++)
            temp_buffer[cnt++] = 0x00;
         // certificate type & algorithm
         temp_buffer[cnt++] = ec->cert_type;
         // continuation pointer
         temp_buffer[cnt++] = ec->pkt_cont_ptr;
         // zero crc 
         temp_buffer[cnt++] = 0x00;
         temp_buffer[cnt++] = 0x00;
         buf_len = 32;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Temporary buffer set to verify account file");
         break;
   
      // set account info in temp buffer (money, transaction ID, MAC)
      case ST_SET_BUFF_ACCNT:
         // save mac 
         if (!prslt)
         {
            sprintf(statusmsg,"eCertificate object not present");
            *Status = STATUS_ERROR_HALT;
            break;
         }
         // get a pointer to the eCert
         ec = (eCertificate*)prslt;
         // length byte
         cnt = 0;
         temp_buffer[cnt++] = 0x1D;
         // transaction ID
         temp_buffer[cnt++] = ec->trans_id & 0xFF;
         temp_buffer[cnt++] = (ec->trans_id >> 8) & 0xFF;
         // monetary unit code & Multiplier
         temp_buffer[cnt++] = ec->code_mult & 0xFF;
         temp_buffer[cnt++] = (ec->code_mult >> 8) & 0xFF;
         // money balance 
         for (i = 0; i < 3; i++)
            temp_buffer[cnt++] = ec->balance[i];
         // money MAC 
         for (i = 0; i < 20; i++)
            temp_buffer[cnt++] = ec->money_mac[i];
         // certificate type & algorithm
         temp_buffer[cnt++] = ec->cert_type;
         // continuation pointer
         temp_buffer[cnt++] = ec->pkt_cont_ptr;
         // calculate crc if MAC is not all zeros
         CRC16 = ec->money_page;
         // calculate crc16
         for (i = 0; i < cnt; i++)
            docrc16(temp_buffer[i]);
         // add the crc16
         temp_buffer[cnt++] = (uchar)(~(CRC16 & 0xFF));
         temp_buffer[cnt++] = (uchar)(~((CRC16 & 0xFF00) >> 8));
         buf_len = 32;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Temporary buffer set to account file");
         break;

      // increment current page counter
      case ST_INC_COUNTER:
         if (prslt)
         {
            ec = (eCertificate*)prslt;
            ec->write_cycle_cnt++;
         }
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Page write counter incremented");
         break;

      // set counter/ page number/ rom / random value in temp buffer
      case ST_SET_BUFF_VAL:
         // save mac 
         if (!prslt)
         {
            sprintf(statusmsg,"eCertificate object not present");
            *Status = STATUS_ERROR_HALT;
            break;
         }
         // get a pointer to the eCert
         ec = (eCertificate*)prslt;
         // first set all to FF
         for (i = 0; i < 32; i++)
            temp_buffer[i] = 0xFF;
         // write page counter
         tcnt = ec->write_cycle_cnt;
         for (i = 8; i <= 11; i++) 
         {
            temp_buffer[i] = (uchar)(tcnt & 0xFF);
            tcnt >>= 8;
         }
         // page number 
         temp_buffer[12] = ec->money_page;
         // rom number of token
         for (i = 0; i < 7; i++)
            temp_buffer[i+13] = ec->serial_num[i];   
         // random challenge
         for (i = 0; i < 3; i++)
            temp_buffer[i+20] = randnum[i]; 
         buf_len = 32;
         *Status = STATUS_STEP_COMPLETE;
         sprintf(statusmsg,"Temporary buffer set with validate data");
         break;         

      // set temp buffer from data in prslt
      case ST_BUF_INPUT:
         // check if have a pointer to something
         if (!prslt)
         {
            sprintf(statusmsg,"Input buffer object not present");
            *Status = STATUS_ERROR_HALT;
            break;
         }
         // get a pointer to the buffer
         puchar = (uchar *)prslt;
         for (i = 0; i < rslt_max; i++)
            temp_buffer[i] = puchar[i];
         buf_len = rslt_max;
         sprintf(statusmsg,"Input data copied to temporary buffer");
         *Status = STATUS_STEP_COMPLETE;
         break;

      // format temp buffer to UDP 
      case ST_FORMAT_UDP:
         // extract page number (address) and shift data
         address = temp_buffer[0] << 5;
         for (i = 1; i <= buf_len; i++)
            temp_buffer[i - 1] = temp_buffer[i];
         buf_len--;
         if (FormatUDP(address,temp_buffer,&buf_len))
         {
            // change the length to write the entire page (to use CRC16)
            buf_len = 32;
            sprintf(statusmsg,"Temporary buffer no UDP format");
            *Status = STATUS_STEP_COMPLETE;
         }
         else   
         {
            sprintf(statusmsg,"Error converting to UDP (too long)");
            *Status = STATUS_ERROR_HALT;
         }
         break;

      default:
         *Status = STATUS_ERROR_HALT;
         sprintf(statusmsg,"1-Wire Step not known\n");
         break;
   }

   return *Status;
}


//----------------------------------------------------------------------
// Verify that this is the correct format for a money page (UDP). Copy
// from RAW ecert to the ecert structure
//
// Return: TRUE - read sucessfull
//         FALSE - wrong format
//                                                           
int Make_eCertificate(ushort address, uchar *buf, eCertificate *ec)
{
   int i,cnt;
   short ROM[8];

   // verify length
   if (buf[0] != 0x1D)
      return FALSE;

   // verify the crc16 
   CRC16 = address >> 5;
   for (i = 0; i < buf[0] + 3; i++)
      docrc16(buf[i]);
      
   // verify CRC16 is correct
   if (CRC16 != 0xB001)
      return FALSE;

   // copy the contents over into the eCertificate structure
   ec->obj_descriptor = 0x0501;
   ec->obj_len = sizeof(ec) - 3;
   for (i = 0; i < 20; i++)
      ec->constant_data[i] = 0;
   cnt = 0;
   ec->pkt_len = buf[cnt++];
   ec->trans_id = buf[cnt] | (buf[cnt+1] << 8);
   cnt += 2;
   ec->code_mult = buf[cnt] | (buf[cnt+1] << 8);
   cnt += 2;
   for (i = 0; i < 3; i++)
      ec->balance[i] = buf[cnt++];
   for (i = 0; i < 20; i++)
      ec->money_mac[i] = buf[cnt++];
   ec->cert_type = buf[cnt++];
   ec->pkt_cont_ptr = buf[cnt++];
   for (i = 0; i < 2; i++)
      ec->inv_crc[i] = buf[cnt++];
   ec->write_cycle_cnt = 0;
   for (i = 3; i >= 0; i--)
   {
      ec->write_cycle_cnt <<= 8;
      ec->write_cycle_cnt |= buf[cnt + i];
   }      
   cnt += 4;
   ec->money_page = address >> 5;

   ROM[0] = 0;
   TMRom(session_handle[dev],&state_buffer[dev][0],ROM);
   for (i = 0; i < 7; i++)
      ec->serial_num[i] = (uchar)ROM[i];

   // get user data from global
   sprintf(ec->user_data,"%s",user_data);

   return TRUE;
}