//---------------------------------------------------------------------------
// Copyright (C) 2000 Dallas Semiconductor Corporation, All Rights Reserved.
// 
// Permission is hereby granted, free of charge, to any person obtaining a 
// copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included 
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES 
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
// OTHER DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of Dallas Semiconductor 
// shall not be used except as stated in the Dallas Semiconductor 
// Branding Policy. 
//---------------------------------------------------------------------------
//
//  ibshaut.c - Utility functions for SHA iButton applications
//
//  Compiler:  Microsoft Visual C 
//  Version:   1.02
//  History:   
//             1.01 -> 1.02  Used correct MUC (Monetary Units Code) and 
//                           multiplier for US dollars
//

#include "ibtmexcw.h"   
#include "ibsha.h"

// external programs
extern int RunScriptSHA(Script script[], void *, int);
extern int ReadUDP_SHA(ushort address, uchar *, int *);

// local function prototypes
int PrintHeader(char *);
void output_status(int, char *);
int EnterString(char *, char *, int, int);
int EnterNum(char *, int, long *, long, long);
int ToHex(char ch);
int GetSeviceProviderSettings(int, char *, uchar *, int, uchar *);
int GetRovingSettings(uchar *, ulong *);
int WriteFileSHA(uchar *, uchar *, int, uchar *);
int bitacc(int, int, int, uchar *);
void ExitProg(char *, int);
int ReadFileSHA(uchar *, int *, uchar *, int *, int);

//--------------------------------------------------------------------------
// Prints a message, and the current driver versions.  Return TRUE
// driver version were found and FALSE if not.
//
int PrintHeader(char *title)
{
   char tstr[80], msg[512]; 
   int msgcnt = 0,rt=FALSE;

   // print a header with driver version info   
   sprintf(msg,"\n%s\n",title);
   output_status(LV_ALWAYS,msg);
   output_status(LV_ALWAYS,"Dallas Semiconductor Corporation\n");
   output_status(LV_ALWAYS,"Copyright (C) 1992-2000\n\n");
   sprintf(msg,"Driver Info:\n   Port number: %d    Port type: %d\n",PortNum[dev],PortType[dev]);
   output_status(LV_ALWAYS,msg);
   Get_Version(tstr);
   sprintf(msg,"   Main driver version: %s\n",tstr);
   output_status(LV_ALWAYS,msg);
   msgcnt = sprintf(msg + msgcnt,"   Sub-driver version, Type%d:",PortType[dev]);
   if (TMGetTypeVersion(PortType[dev],tstr) < 0)
      msgcnt += sprintf(msg + msgcnt,"No Hardware Driver for this type found!");
   else
   {
      rt = TRUE;
      msgcnt += sprintf(msg + msgcnt," %s\n\n",tstr);
   }
   // send output
   output_status(LV_ALWAYS,msg);

   return rt;
}


//--------------------------------------------------------------------------
//  output status message
//
void output_status(int level, char *st)
{
   // skip null strings
   if (st[0] == 0)
      return;

   // check if in verbose mode
   if ((level >= LV_ALWAYS) || VERBOSE)
   { 
      printf("%s",st);
      if(fp != NULL)
         fprintf(fp,"%s",st);
   }
}

//----------------------------------------------------------------------
// Get the settings for the Service Provider
//
int GetSeviceProviderSettings(int get_sp, char *sp_name, uchar *auth_secret, int get_money, uchar *money_secret)
{
   char temp[255];
   int i,len,cnt;

   // instructions
   output_status(LV_ALWAYS,"Default values specified in ()\n"
                "Hit Escape to abort\n");  

   // service provider name
   if (get_sp)
   {
      sprintf(sp_name,"DLSM");
      len = EnterString("\nEnter the service provider name",sp_name,1,4);
      if (len < 0)
         return FALSE;
      else
      {
         // file with spaces to make easier on compare
         for (i = len; i < 4; i++)
            sp_name[i] = ' ';
         sp_name[4] = 0;  
         sprintf(temp,"Service provider name selected: %s\n",sp_name);
         output_status(LV_ALWAYS,temp);  
      }
   }

   // auth secret
   sprintf(temp,"FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF");
   len = EnterString("\nEnter service provider AUTHENTICATE secret (hex)\n",temp,2,64);
   if (len < 0)
      return FALSE;
   else
   {
      // convert to byte array with FF padding
      for (i = 0; i < len; i += 2)
      {
         auth_secret[i / 2] = ToHex(temp[i]) << 4;
         auth_secret[i / 2] |= ToHex(temp[i + 1]);
      }     
      for (i = len; i < 64; i += 2)
         auth_secret[i / 2] = 0xFF;

      cnt = sprintf(temp,"Service provider AUTHENTICATE secret entered (hex):\n ");
      for (i = 0; i < 32; i++)
         cnt += sprintf(temp + cnt,"%02X",auth_secret[i]);
      cnt += sprintf(temp + cnt,"\n");
      output_status(LV_ALWAYS,temp);  
   }

   // money secret
   if (get_money)
   {
      sprintf(temp,"FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF");
      len = EnterString("\nEnter service provider MONEY secret (hex)\n",temp,2,64);
      if (len < 0)
         return FALSE;
      else
      {
         // convert to byte array with FF padding
         for (i = 0; i < len; i += 2)
         {
            money_secret[i / 2] = ToHex(temp[i]) << 4;
            money_secret[i / 2] |= ToHex(temp[i + 1]);
         }     
         for (i = len; i < 64; i += 2)
            money_secret[i / 2] = 0xFF;

         cnt = sprintf(temp,"Service provider MONEY secret entered (hex):\n ");
         for (i = 0; i < 32; i++)
            cnt += sprintf(temp + cnt,"%02X",money_secret[i]);
         cnt += sprintf(temp + cnt,"\n");
         output_status(LV_ALWAYS,temp);  
      }
   }

   return TRUE;
}


//----------------------------------------------------------------------
// Get the settings for the Service Provider
//
int GetRovingSettings(uchar *rov_information, ulong *money)
{
   char temp[255];
   int i,len;

   // user information
   rov_information[29] = 0;
   len = EnterString("\nEnter roving user fixed information\n",rov_information,0,28);
   if (len < 0)
      return FALSE;
   else
   {
      rov_information[len] = 0;
      sprintf(temp,"Roving user fixed information: %s\n",rov_information);
      output_status(LV_ALWAYS,temp);  

      // pad the user information with spaces up to 28 characters
      for (i = len; i < 28; i++)
         rov_information[i] = ' ';
      // pad rest with zeros (money page)
      for (i = 28; i < 56; i++)
         rov_information[i] = 0;
   }

   // money (1.02) change to dollars
   if (!EnterNum("\nEnter the starting money balance (in US dollars)\n", 5, money, 0, 99999))
      return FALSE;
   else
   {
      sprintf(temp,"Starting balance: $%d US\n",*money);
      output_status(LV_ALWAYS,temp);  
   }

   return TRUE;
}


//----------------------------------------------------------------------
// Enter a string 
// Leave the same if no characters entered
//                                                          
int EnterString(char *msg, char *buf, int min, int max)
{
   int len,deflen;
   char ch,defbuf[80];
   
   // remember the start length
   deflen = strlen(buf);
   if (deflen < 80)
      sprintf(defbuf,buf);
   else
      defbuf[0] = 0;

   // prompt
   MessageBeep(MB_OK);
   if (deflen < 30)
      printf("%s (%s): ",msg,buf);
   else
      printf("%s (%s):\n",msg,buf);

   // loop to get the file 
   len = 0;
   for (;;)
   {
      // get a key 
      ch = getch();
      // check for special character (non-ascii)
      if (kbhit())
      {
         ch = getch();
         continue;  // skip it   
      }

      // check the backspace key 
      if (ch == 0x08)
      {
         if (len > 0)
         {
            // got a valid backspace 
            len--;
            printf("\x008 \x008");
         }
      }
      // escape key
      else if (ch == 0x1B)
      {
         printf("\n");
         return -1; 
      }
      // caraige return
      else if (ch == '\r')
      {
         // check for special case (keep default)
         if ((len == 0) && (deflen > min)) 
         {
            sprintf(buf,"%s",defbuf);
            printf("\n");
            return deflen;
         }
         else if (len < min)
            continue;
         else
         {
            printf("\n");
            return len;
         }
      }
      // valid key 
      else if (len < max)
      {
         printf("%c",ch);
         // record the char 
         buf[len++] = ch;
      }
   }
}


//----------------------------------------------------------------------
// Enter a decimal string and convert it to an unsigned long
// Prompt again if not within min/max inclusive.
//                                                          
int EnterNum(char *msg, int numchars, long *value, long min, long max)
{
#ifdef WIN32 
   short tmp,cnt,isneg=FALSE; 
   char ch;

   // loop while not in correct range
   do
   {
      MessageBeep(MB_OK);
      printf("%s (%d): ",msg,*value);

      // loop for each character read
      cnt = 0;
      for (;;)
      {
         ch = getch();

         // negative flag
         if (ch == '-')
         {
            if (!isneg)
            {
               isneg = TRUE;
               printf("-");
               cnt++;
            }
         }
         // backspace 
         if (ch == 0x08)
         {
            if (cnt)
            {
               if (isneg && (cnt == 1))
                  isneg = FALSE;
               else
                  *value /= 10;
               printf("%c %c",ch,ch);
               cnt--;
            } 
         }
         // escape 
         if (ch == 0x1B)
         {
            printf("  Aborted\n\n");
            return FALSE;
         }
         // enter 
         if (ch == 0x0D)
         {
            printf("\n");
            break;
         }
         // number 
         else if ((ch >= '0') && (ch <= '9'))
         {
            if (cnt == 0)
               *value = 0;

            if (cnt < numchars)
            {
               printf("%c",ch);
               tmp = ch - 0x30;
               *value *= 10; 
               *value += tmp;
               cnt++;
            }
         }
      }

      if (isneg)
         *value = -*value;
   }
   while ((*value < min) || (*value > max));

   return TRUE;
#else
   char line[256];

   // loop while not in correct range
   do
   {
      printf("%s",msg);
      *value = 0;

      // read a line
      gets(line);

      // attempt to convert to a integer
      *value = atoi(line);
   }
   while ((*value < min) || (*value > max));

   return TRUE;
#endif
}


//------------------------------------------------------------------------
// Write the file 'filename' to the current device with the provided data.
// The pages used the file will be specified in the 'page_map' array.
// Each element in the array represents a page number (0 to 255).
// For example if the page data is 2 pages worth, page_am could contain
// an array (0x04,0x12) to specify page 4 and page 12.  
//
// *** Current version of this function will FORMAT the device before
//     writing the file!
//
int WriteFileSHA(uchar *filename, uchar *buf, int fllen, uchar *page_map)
{
   uchar pgbuf[32];
   uchar newdir[] = { 0, 0xAA, 0, 0x80, 0x01, 0, 0, 0, ' ', ' ', ' ', ' ', 0, 1, 1, 0 };
   int numpgs,flpg,i,cntleft,pos;

   // calculate the number of pages needed to write the file
   numpgs = (fllen / 28) + ((fllen % 28) ? 1 : 0);
   
   // put the file in the newdirectory
   for(i = 0; i < 5; i++)
      newdir[i+8] = filename[i];
   newdir[13] = page_map[0];
   newdir[14] = (uchar)numpgs;

   // set the bitmap in the directory page
   for (i = 0; i < numpgs; i++)
      bitacc(1,1,page_map[i],&newdir[4]);

   // loop to write the file 
   cntleft = fllen;  // count of bytes left to write  
   pos = 0; // current position in the buffer to write
   flpg = 0;
   while (cntleft > 0)
   {
      // set the page number in the first byte (for script)
      pgbuf[0] = page_map[flpg];

      // get a page of data to write
      for (i = 0; i < ((cntleft > 28) ? 28 : cntleft); i++)
         pgbuf[i + 1] = buf[pos++];
                                    
      // adjust the bytes left
      cntleft -= i;
                                    
      // set the next page pointer   
      pgbuf[i + 1] = (cntleft == 0) ? 0 : page_map[flpg+1];
      
      // write the page and check to result
      if (!RunScriptSHA(WritePage, &pgbuf[0], i + 2))
         return FALSE;
      
      // set the next page
      flpg++;      
   } 

   // now write the directory page
   if (!RunScriptSHA(WritePage, &newdir[0], 16))
      return FALSE;

   return TRUE;
}

//------------------------------------------------------------------------
// Convert 1 hex character to binary.  If not a hex character then
// return 0.
//
int ToHex(char ch)
{
   if ((ch >= '0') && (ch <= '9'))
      return ch - 0x30;
   else if ((ch >= 'A') && (ch <= 'F'))
      return ch - 0x37;
   else if ((ch >= 'a') && (ch <= 'f'))
      return ch - 0x57;
   else
      return 0;
}

//------------------------------------------------------------------------
// Print message and exit program
//
void ExitProg(char *msg, int exit_code)
{
   printf("%s\n",msg);
   exit(exit_code);
}

//--------------------------------------------------------------------------
//  Bit utility to read and write a bit in the buffer 'buf'.
//
int bitacc(int op, int state, int loc, uchar *buf)
{
     int nbyt,nbit;

     nbyt = (loc / 8);
     nbit = loc - (nbyt * 8);

     if (op == 1)
     {
          if (state)
             buf[nbyt] |= (0x01 << nbit);
          else
             buf[nbyt] &= ~(0x01 << nbit);

          return 1;
     }
     else
          return ((buf[nbyt] >> nbit) & 0x01);
}


//--------------------------------------------------------------------------
//  Read file specified by 5 byte file name 'filename' and place in
//  'read_buf'.
//
int ReadFileSHA(uchar *filename, int *start_pg, uchar *read_buf, int *read_len, int max_read)
{
   uchar dirpg=0,pgbuf[32],filepg=0;
   int bufcnt=0,i,pglen;
               
   // set the intial read length to 0
   *read_len = 0;

   // loop read directory pages until the file entry is found
   do
   {
      // read a directory page
      if (!ReadUDP_SHA((ushort)(dirpg << 5), pgbuf, &pglen))
         return FALSE;
         
      // if this is the first page make sure this is a directory
      // structure
      if (  ((dirpg == 0) && 
            ((pgbuf[0] != 0xAA) || (pgbuf[1] != 0) || (pglen < 7)))
            ||
            ((pglen-1) % 7) ) 
         return FALSE;
      
      // loop through each file entry in directory page (page 0 exception)
      for (i = (dirpg == 0) ? 7 : 0; i < 28; i += 7)
      {     
         // file entry found?
         if ((filename[0] == pgbuf[i]) &&
             (filename[1] == pgbuf[i+1]) &&
             (filename[2] == pgbuf[i+2]) &&
             (filename[3] == pgbuf[i+3]) &&
             (filename[4] == (pgbuf[i+4] & 0x7F)) )
         {
            // get the file starting page number
            filepg = pgbuf[i+5];
            *start_pg = (int)filepg;
            break;
         }
      }
      
      // get the next directory page (from page pointer) 
      dirpg = pgbuf[pglen-1];
   }
   while (dirpg && (filepg == 0));  
   
   // check if file found
   if (!filepg)
      return FALSE;
   
   // loop to read the file pages
   do
   {
      // read a file page
      if (!ReadUDP_SHA((ushort)(filepg << 5), pgbuf, &pglen))
         return FALSE;
      
      // check for too big
      if ((pglen + bufcnt - 1) > max_read)
         return FALSE;

      // append the page data to the buffer
      for (i = 0; i < (pglen - 1); i++)
         read_buf[bufcnt++] = pgbuf[i];

      // get the next file page (from page pointer) 
      filepg = pgbuf[pglen-1];
   }
   while (filepg);

   // return the number of data bytes read
   *read_len = bufcnt;
   
   return TRUE;
}


